@@ -1,13 +1,17 @@
 const Clutter = imports.gi.Clutter;
+const GLib = imports.gi.GLib;
+const Gio = imports.gi.Gio;
 const Gtk = imports.gi.Gtk;
 const Meta = imports.gi.Meta;
 const Shell = imports.gi.Shell;
 const St = imports.gi.St;
 
+const DND = imports.ui.dnd;
 const Hash = imports.misc.hash;
 const Lang = imports.lang;
 const Main = imports.ui.main;
 const MessageTray = imports.ui.messageTray;
+const PanelMenu = imports.ui.panelMenu;
 const PopupMenu = imports.ui.popupMenu;
 
 const ExtensionUtils = imports.misc.extensionUtils;
@@ -15,10 +19,12 @@ const Me = ExtensionUtils.getCurrentExtension();
 const Convenience = Me.imports.convenience;
 
 const ICON_TEXTURE_SIZE = 24;
+const DND_ACTIVATE_TIMEOUT = 500;
 
 const GroupingMode = {
     NEVER: 0,
-    ALWAYS: 1
+    AUTO: 1,
+    ALWAYS: 2
 };
 
 
@@ -46,6 +52,12 @@ const WindowTitle = new Lang.Class({
         this._label = new St.Label();
         this.actor.add(this._label);
 
+        this._textureCache = St.TextureCache.get_default();
+        this._iconThemeChangedId =
+            this._textureCache.connect('icon-theme-changed', Lang.bind(this,
+                function() {
+                    this._icon.child = app.create_icon_texture(ICON_TEXTURE_SIZE);
+                }));
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
         this._notifyTitleId =
@@ -70,6 +82,7 @@ const WindowTitle = new Lang.Class({
     },
 
     _onDestroy: function() {
+        this._textureCache.disconnect(this._iconThemeChangedId);
         this._metaWindow.disconnect(this._notifyTitleId);
         this._metaWindow.disconnect(this._notifyMinimizedId);
     }
@@ -165,9 +178,9 @@ const AppButton = new Lang.Class({
         this._multiWindowTitle = new St.BoxLayout({ x_expand: true });
         stack.add_actor(this._multiWindowTitle);
 
-        let icon = new St.Bin({ style_class: 'window-button-icon',
-                                child: app.create_icon_texture(ICON_TEXTURE_SIZE) });
-        this._multiWindowTitle.add(icon);
+        this._icon = new St.Bin({ style_class: 'window-button-icon',
+                                  child: app.create_icon_texture(ICON_TEXTURE_SIZE) });
+        this._multiWindowTitle.add(this._icon);
         this._multiWindowTitle.add(new St.Label({ text: app.get_name() }));
 
         this._menuManager = new PopupMenu.PopupMenuManager(this);
@@ -177,6 +190,12 @@ const AppButton = new Lang.Class({
         this._menuManager.addMenu(this._menu);
         Main.uiGroup.add_actor(this._menu.actor);
 
+        this._textureCache = St.TextureCache.get_default();
+        this._iconThemeChangedId =
+            this._textureCache.connect('icon-theme-changed', Lang.bind(this,
+                function() {
+                    this._icon.child = app.create_icon_texture(ICON_TEXTURE_SIZE);
+                }));
         this.actor.connect('clicked', Lang.bind(this, this._onClicked));
         this.actor.connect('destroy', Lang.bind(this, this._onDestroy));
 
@@ -236,11 +255,13 @@ const AppButton = new Lang.Class({
 
         if (this._singleWindowTitle.visible) {
             if (!this._windowTitle) {
-                this._windowTitle = new WindowTitle(windows[0]);
+                this.metaWindow = windows[0];
+                this._windowTitle = new WindowTitle(this.metaWindow);
                 this._singleWindowTitle.child = this._windowTitle.actor;
             }
         } else {
             if (this._windowTitle) {
+                this.metaWindow = null;
                 this._singleWindowTitle.child = null;
                 this._windowTitle = null;
             }
@@ -279,6 +300,7 @@ const AppButton = new Lang.Class({
     },
 
     _onDestroy: function() {
+        this._textureCache.disconnect(this._iconThemeChangedId);
         global.window_manager.disconnect(this._switchWorkspaceId);
         this._windowTracker.disconnect(this._notifyFocusId);
         this.app.disconnect(this._windowsChangedId);
@@ -352,6 +374,106 @@ const TrayButton = new Lang.Class({
     }
 });
 
+const WorkspaceIndicator = new Lang.Class({
+    Name: 'WindowList.WorkspaceIndicator',
+    Extends: PanelMenu.Button,
+
+    _init: function(){
+        this.parent(0.0, _("Workspace Indicator"));
+        this.actor.add_style_class_name('window-list-workspace-indicator');
+
+        this._currentWorkspace = global.screen.get_active_workspace().index();
+        this.statusLabel = new St.Label({ text: this._getStatusText() });
+
+        this.actor.add_actor(this.statusLabel);
+
+        this.workspacesItems = [];
+
+        this._screenSignals = [];
+        this._screenSignals.push(global.screen.connect_after('workspace-added', Lang.bind(this,this._updateMenu)));
+        this._screenSignals.push(global.screen.connect_after('workspace-removed', Lang.bind(this,this._updateMenu)));
+        this._screenSignals.push(global.screen.connect_after('workspace-switched', Lang.bind(this,this._updateIndicator)));
+
+        this.actor.connect('scroll-event', Lang.bind(this, this._onScrollEvent));
+        this._updateMenu();
+
+        this._settings = new Gio.Settings({ schema: 'org.gnome.desktop.wm.preferences' });
+        this._settingsChangedId = this._settings.connect('changed::workspace-names', Lang.bind(this, this._updateMenu));
+    },
+
+    destroy: function() {
+        for (let i = 0; i < this._screenSignals.length; i++)
+            global.screen.disconnect(this._screenSignals[i]);
+
+        if (this._settingsChangedId) {
+            this._settings.disconnect(this._settingsChangedId);
+            this._settingsChangedId = 0;
+        }
+
+        this.parent();
+    },
+
+    _updateIndicator: function() {
+        this.workspacesItems[this._currentWorkspace].setShowDot(false);
+        this._currentWorkspace = global.screen.get_active_workspace().index();
+        this.workspacesItems[this._currentWorkspace].setShowDot(true);
+
+        this.statusLabel.set_text(this._getStatusText());
+    },
+
+    _getStatusText: function() {
+        let current = global.screen.get_active_workspace().index();
+        let total = global.screen.n_workspaces;
+
+        return '%d / %d'.format(current + 1, total);
+    },
+
+    _updateMenu: function() {
+        this.menu.removeAll();
+        this.workspacesItems = [];
+        this._currentWorkspace = global.screen.get_active_workspace().index();
+
+        for(let i = 0; i < global.screen.n_workspaces; i++) {
+            let name = Meta.prefs_get_workspace_name(i);
+            let item = new PopupMenu.PopupMenuItem(name);
+            item.workspaceId = i;
+
+            item.connect('activate', Lang.bind(this, function(item, event) {
+                this._activate(item.workspaceId);
+            }));
+
+            if (i == this._currentWorkspace)
+                item.setShowDot(true);
+
+            this.menu.addMenuItem(item);
+            this.workspacesItems[i] = item;
+        }
+
+        this.statusLabel.set_text(this._getStatusText());
+    },
+
+    _activate: function(index) {
+        if(index >= 0 && index < global.screen.n_workspaces) {
+            let metaWorkspace = global.screen.get_workspace_by_index(index);
+            metaWorkspace.activate(global.get_current_time());
+        }
+    },
+
+    _onScrollEvent: function(actor, event) {
+        let direction = event.get_scroll_direction();
+        let diff = 0;
+        if (direction == Clutter.ScrollDirection.DOWN) {
+            diff = 1;
+        } else if (direction == Clutter.ScrollDirection.UP) {
+            diff = -1;
+        } else {
+            return;
+        }
+
+        let newIndex = this._currentWorkspace + diff;
+        this._activate(newIndex);
+    },
+});
 
 const WindowList = new Lang.Class({
     Name: 'WindowList',
@@ -381,6 +503,26 @@ const WindowList = new Lang.Class({
                 let spacing = node.get_length('spacing');
                 this._windowList.layout_manager.spacing = spacing;
             }));
+        this._windowList.connect('notify::allocation', Lang.bind(this,
+            function() {
+                if (this._groupingMode != GroupingMode.AUTO || this._grouped)
+                    return;
+
+                let allocation = this._windowList.allocation;
+                let width = allocation.x2 - allocation.x1;
+                let [, natWidth] = this._windowList.get_preferred_width(-1);
+                if (width < natWidth) {
+                    this._grouped = true;
+                    Meta.later_add(Meta.LaterType.BEFORE_REDRAW,
+                                   Lang.bind(this, this._populateWindowList));
+                }
+            }));
+
+        this._workspaceIndicator = new WorkspaceIndicator();
+        box.add(this._workspaceIndicator.container);
+
+        this._menuManager = new PopupMenu.PopupMenuManager(this);
+        this._menuManager.addMenu(this._workspaceIndicator.menu);
 
         this._trayButton = new TrayButton();
         box.add(this._trayButton.actor);
@@ -423,12 +565,30 @@ const WindowList = new Lang.Class({
 
         this._overviewHidingId =
             Main.overview.connect('hiding', Lang.bind(this, function() {
-                this.actor.show();
+                this.actor.visible = !Main.layoutManager.primaryMonitor.inFullscreen;
                 this._updateKeyboardAnchor();
                 this._updateMessageTrayAnchor();
             }));
         this._updateMessageTrayAnchor();
 
+        this._fullscreenChangedId =
+            global.screen.connect('in-fullscreen-changed', Lang.bind(this, function() {
+                this._updateMessageTrayAnchor();
+            }));
+
+        this._dragBeginId =
+            Main.xdndHandler.connect('drag-begin',
+                                     Lang.bind(this, this._onDragBegin));
+        this._dragEndId =
+            Main.xdndHandler.connect('drag-end',
+                                     Lang.bind(this, this._onDragEnd));
+        this._dragMonitor = {
+            dragMotion: Lang.bind(this, this._onDragMotion)
+        };
+
+        this._dndTimeoutId = 0;
+        this._dndWindow = null;
+
         this._settings = Convenience.getSettings();
         this._groupingModeChangedId =
             this._settings.connect('changed::grouping-mode',
@@ -438,13 +598,14 @@ const WindowList = new Lang.Class({
 
     _groupingModeChanged: function() {
         this._groupingMode = this._settings.get_enum('grouping-mode');
+        this._grouped = this._groupingMode == GroupingMode.ALWAYS;
         this._populateWindowList();
     },
 
     _populateWindowList: function() {
         this._windowList.destroy_all_children();
 
-        if (this._groupingMode == GroupingMode.NEVER) {
+        if (!this._grouped) {
             let windows = Meta.get_window_actors(global.screen);
             for (let i = 0; i < windows.length; i++)
                 this._onWindowAdded(null, windows[i].metaWindow);
@@ -477,7 +638,7 @@ const WindowList = new Lang.Class({
     },
 
     _onAppStateChanged: function(appSys, app) {
-        if (this._groupingMode != GroupingMode.ALWAYS)
+        if (!this._grouped)
             return;
 
         if (app.state == Shell.AppState.RUNNING)
@@ -508,7 +669,7 @@ const WindowList = new Lang.Class({
         if (!Shell.WindowTracker.get_default().is_window_interesting(win))
             return;
 
-        if (this._groupingMode != GroupingMode.NEVER)
+        if (this._grouped)
             return;
 
         let button = new WindowButton(win);
@@ -519,8 +680,14 @@ const WindowList = new Lang.Class({
     },
 
     _onWindowRemoved: function(ws, win) {
-        if (this._groupingMode != GroupingMode.NEVER)
+        if (this._grouped) {
+            if (this._groupingMode == GroupingMode.AUTO) {
+                this._grouped = false;
+                this._populateWindowList();
+            }
+
             return;
+        }
 
         let children = this._windowList.get_children();
         for (let i = 0; i < children.length; i++) {
@@ -559,7 +726,61 @@ const WindowList = new Lang.Class({
         }
     },
 
+    _onDragBegin: function() {
+        DND.addDragMonitor(this._dragMonitor);
+    },
+
+    _onDragEnd: function() {
+        DND.removeDragMonitor(this._dragMonitor);
+        this._removeActivateTimeout();
+    },
+
+    _onDragMotion: function(dragEvent) {
+        if (Main.overview.visible ||
+            !this.actor.contains(dragEvent.targetActor)) {
+            this._removeActivateTimeout();
+            return DND.DragMotionResult.CONTINUE;
+        }
+
+        let hoveredWindow = null;
+        if (dragEvent.targetActor._delegate)
+            hoveredWindow = dragEvent.targetActor._delegate.metaWindow;
+
+        if (!hoveredWindow ||
+            this._dndWindow == hoveredWindow)
+            return DND.DragMotionResult.CONTINUE;
+
+        this._removeActivateTimeout();
+
+        this._dndWindow = hoveredWindow;
+        this._dndTimeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT,
+                                              DND_ACTIVATE_TIMEOUT,
+                                              Lang.bind(this, this._activateWindow));
+
+        return DND.DragMotionResult.CONTINUE;
+    },
+
+    _removeActivateTimeout: function() {
+        if (this._dndTimeoutId)
+            GLib.source_remove (this._dndTimeoutId);
+        this._dndTimeoutId = 0;
+        this._dndWindow = null;
+    },
+
+    _activateWindow: function() {
+        let [x, y] = global.get_pointer();
+        let pickedActor = global.stage.get_actor_at_pos(Clutter.PickMode.ALL, x, y);
+
+        if (this._dndWindow && this.actor.contains(pickedActor))
+            this._dndWindow.activate(global.get_current_time());
+        this._dndWindow = null;
+        this._dndTimeoutId = 0;
+
+        return false;
+    },
+
     _onDestroy: function() {
+        this._workspaceIndicator.destroy();
 
         Main.ctrlAltTabManager.removeGroup(this.actor);
 
@@ -584,6 +805,8 @@ const WindowList = new Lang.Class({
         Main.overview.disconnect(this._overviewShowingId);
         Main.overview.disconnect(this._overviewHidingId);
 
+        global.screen.disconnect(this._fullscreenChangedId);
+
         this._settings.disconnect(this._groupingModeChangedId);
 
         let windows = Meta.get_window_actors(global.screen);
